@@ -1,6 +1,7 @@
 /* OpenBSC Abis input driver for ip.access */
 
-/* (C) 2009-2021 by Harald Welte <laforge@gnumonks.org>
+/* (C) 2024 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
+ * (C) 2009-2021 by Harald Welte <laforge@gnumonks.org>
  * (C) 2010 by Holger Hans Peter Freyther
  * (C) 2010 by On-Waves
  *
@@ -39,29 +40,26 @@
 #include <arpa/inet.h>
 
 #include <osmocom/core/select.h>
-#include <osmocom/gsm/tlv.h>
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/macaddr.h>
 #include <osmocom/core/logging.h>
 #include <osmocom/core/talloc.h>
-#include <osmocom/abis/e1_input.h>
-#include <osmocom/abis/ipaccess.h>
 #include <osmocom/core/socket.h>
-#include <osmocom/abis/ipa.h>
 #include <osmocom/core/backtrace.h>
-#include <osmocom/gsm/ipa.h>
 #include <osmocom/core/stats_tcp.h>
 #include <osmocom/core/fsm.h>
+#include <osmocom/gsm/tlv.h>
+#include <osmocom/gsm/protocol/ipaccess.h>
+#include <osmocom/gsm/ipa.h>
 #include <osmocom/netif/stream.h>
 #include <osmocom/netif/ipa.h>
+#include <osmocom/abis/e1_input.h>
 
 /* global parameters of IPA input driver */
 struct ipa_pars g_e1inp_ipaccess_pars;
 
 static void *tall_ipa_ctx;
 
-#define TS1_ALLOC_SIZE	900
-
 #define DEFAULT_TCP_KEEPALIVE_IDLE_TIMEOUT 30
 #define DEFAULT_TCP_KEEPALIVE_INTERVAL     3
 #define DEFAULT_TCP_KEEPALIVE_RETRY_COUNT  10
@@ -81,105 +79,58 @@ static inline struct e1inp_ts *ipaccess_line_ts(struct osmo_fd *bfd, struct e1in
 
 static inline void ipaccess_keepalive_fsm_cleanup(struct e1inp_ts *e1i_ts)
 {
-	struct osmo_fsm_inst *ka_fsm;
+	struct osmo_ipa_ka_fsm_inst *ka_fsm;
 
 	ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 	if (ka_fsm) {
-		osmo_fsm_inst_term(ka_fsm, OSMO_FSM_TERM_REQUEST, NULL);
 		e1i_ts->driver.ipaccess.ka_fsm = NULL;
+		osmo_ipa_ka_fsm_stop(ka_fsm);
+		osmo_ipa_ka_fsm_free(ka_fsm);
 	}
 }
 
-static int ipaccess_drop(struct osmo_fd *bfd, struct e1inp_line *line)
+static int ipa_bsc_keepalive_send_cb(struct osmo_ipa_ka_fsm_inst *ka_fi, struct msgb *msg, void *data)
 {
-	int ret = 1;
-	struct e1inp_ts *e1i_ts = ipaccess_line_ts(bfd, line);
-	e1inp_line_get2(line, __func__);
-
-	ipaccess_keepalive_fsm_cleanup(e1i_ts);
-
-	/* Error case: we did not see any ID_RESP yet for this socket. */
-	if (bfd->fd != -1) {
-		LOGPITS(e1i_ts, DLINP, LOGL_NOTICE, "Forcing socket shutdown\n");
-		osmo_fd_unregister(bfd);
-		close(bfd->fd);
-		bfd->fd = -1;
-		switch (line->ops->cfg.ipa.role) {
-		case E1INP_LINE_R_BSC:
-			/* This is BSC code, ipaccess_drop() is only called for
-			   accepted() sockets, hence the bfd holds a reference to
-			   e1inp_line in ->data that needs to be released */
-			OSMO_ASSERT(bfd->data == line);
-			bfd->data = NULL;
-			e1inp_line_put2(line, "ipa_bfd");
-			break;
-		case E1INP_LINE_R_BTS:
-			/* BTS code: bfd->data contains pointer to struct
-			 * ipa_client_conn. Leave it alive so it reconnects.
-			 */
-			break;
-		default:
-			break;
-		}
-		ret = -ENOENT;
-	} else {
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR,
-			"Forcing socket shutdown with no signal link set\n");
-	}
-
-	msgb_free(e1i_ts->pending_msg);
-	e1i_ts->pending_msg = NULL;
-
-	/* e1inp_sign_link_destroy releases the socket descriptors for us. */
-	if (line->ops->sign_link_down)
-		line->ops->sign_link_down(line);
-
-	e1inp_line_put2(line, __func__);
-	return ret;
+	struct osmo_stream_srv *srv = data;
+	osmo_stream_srv_send(srv, msg);
+	return 0;
 }
 
-static void ipa_bsc_keepalive_write_server_cb(struct osmo_fsm_inst *fi, void *conn, struct msgb *msg)
+static int ipa_bsc_keepalive_timeout_cb(struct osmo_ipa_ka_fsm_inst *ka_fi, void *data)
 {
-	struct osmo_fd *bfd = (struct osmo_fd *)conn;
-	write(bfd->fd, msg->data, msg->len);
-	msgb_free(msg);
+	struct osmo_stream_srv *srv = data;
+	osmo_stream_srv_destroy(srv);
+	return 0;
 }
 
-static int ipa_bsc_keepalive_timeout_cb(struct osmo_fsm_inst *fi, void *data)
-{
-	struct osmo_fd *bfd = (struct osmo_fd *)data;
-
-	if (bfd->fd == -1)
-		return 1;
-
-	ipaccess_drop(bfd, (struct e1inp_line *)bfd->data);
-	return 1;
-}
-
-static void ipaccess_bsc_keepalive_fsm_alloc(struct e1inp_ts *e1i_ts, struct osmo_fd *bfd, const char *id)
+static void ipaccess_bsc_keepalive_fsm_alloc(struct e1inp_ts *e1i_ts, struct osmo_stream_srv *conn, const char *id)
 {
 	struct e1inp_line *line = e1i_ts->line;
-	struct osmo_fsm_inst *ka_fsm;
+	struct osmo_ipa_ka_fsm_inst *ka_fsm;
 
 	ipaccess_keepalive_fsm_cleanup(e1i_ts);
 	if (!line->ipa_kap)
 		return;
 
-	ka_fsm = ipa_generic_conn_alloc_keepalive_fsm(tall_ipa_ctx, bfd, line->ipa_kap, id);
+	ka_fsm = osmo_ipa_ka_fsm_alloc(conn, id);
 	e1i_ts->driver.ipaccess.ka_fsm = ka_fsm;
 	if (!ka_fsm) {
 		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Failed to allocate IPA keepalive FSM\n");
 		return;
 	}
-
-	ipa_keepalive_fsm_set_timeout_cb(ka_fsm, ipa_bsc_keepalive_timeout_cb);
-	ipa_keepalive_fsm_set_send_cb(ka_fsm, ipa_bsc_keepalive_write_server_cb);
-	ipa_keepalive_fsm_start(ka_fsm);
+	osmo_ipa_ka_fsm_set_ping_interval(ka_fsm, line->ipa_kap->interval);
+	osmo_ipa_ka_fsm_set_pong_timeout(ka_fsm, line->ipa_kap->wait_for_resp);
+	osmo_ipa_ka_fsm_set_data(ka_fsm, conn);
+	osmo_ipa_ka_fsm_set_send_cb(ka_fsm, ipa_bsc_keepalive_send_cb);
+	osmo_ipa_ka_fsm_set_timeout_cb(ka_fsm, ipa_bsc_keepalive_timeout_cb);
+	osmo_ipa_ka_fsm_start(ka_fsm);
 }
 
-static void ipa_bts_keepalive_write_client_cb(struct osmo_fsm_inst *fi, void *conn, struct msgb *msg) {
-	struct osmo_stream_cli *cli = (struct osmo_stream_cli *)conn;
+static int ipa_bts_keepalive_send_cb(struct osmo_ipa_ka_fsm_inst *ka_fi, struct msgb *msg, void *data)
+{
+	struct osmo_stream_cli *cli = data;
 	osmo_stream_cli_send(cli, msg);
+	return 0;
 }
 
 static void _ipaccess_bts_down_cb(struct osmo_stream_cli *cli)
@@ -192,30 +143,33 @@ static void _ipaccess_bts_down_cb(struct osmo_stream_cli *cli)
 		line->ops->sign_link_down(line);
 }
 
-static int ipa_bts_keepalive_timeout_cb(struct osmo_fsm_inst *fi, void *conn) {
-	struct osmo_stream_cli *cli = (struct osmo_stream_cli *)conn;
+static int ipa_bts_keepalive_timeout_cb(struct osmo_ipa_ka_fsm_inst *ka_fi, void *data)
+{
+	struct osmo_stream_cli *cli = data;
 	_ipaccess_bts_down_cb(cli);
-	return 1;
+	return 0;
 }
 
 static void ipaccess_bts_keepalive_fsm_alloc(struct e1inp_ts *e1i_ts, struct osmo_stream_cli *client, const char *id)
 {
 	struct e1inp_line *line = e1i_ts->line;
-	struct osmo_fsm_inst *ka_fsm;
+	struct osmo_ipa_ka_fsm_inst *ka_fsm;
 
 	OSMO_ASSERT(e1i_ts->driver.ipaccess.ka_fsm == NULL);
 	if (!line->ipa_kap)
 		return;
 
-	ka_fsm = ipa_generic_conn_alloc_keepalive_fsm(client, client, line->ipa_kap, id);
+	ka_fsm = osmo_ipa_ka_fsm_alloc(client, id);
 	e1i_ts->driver.ipaccess.ka_fsm = ka_fsm;
 	if (!ka_fsm) {
 		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Failed to allocate IPA keepalive FSM\n");
 		return;
 	}
-
-	ipa_keepalive_fsm_set_timeout_cb(ka_fsm, ipa_bts_keepalive_timeout_cb);
-	ipa_keepalive_fsm_set_send_cb(ka_fsm, ipa_bts_keepalive_write_client_cb);
+	osmo_ipa_ka_fsm_set_ping_interval(ka_fsm, line->ipa_kap->interval);
+	osmo_ipa_ka_fsm_set_pong_timeout(ka_fsm, line->ipa_kap->wait_for_resp);
+	osmo_ipa_ka_fsm_set_data(ka_fsm, client);
+	osmo_ipa_ka_fsm_set_send_cb(ka_fsm, ipa_bts_keepalive_send_cb);
+	osmo_ipa_ka_fsm_set_timeout_cb(ka_fsm, ipa_bts_keepalive_timeout_cb);
 }
 
 /* See how ts->num is assigned in e1inp_line_create: line->ts[i].num = i+1;
@@ -250,25 +204,40 @@ static inline struct osmo_stream_cli *ipaccess_bts_e1i_ts_stream_cli(const struc
 	return cli;
 }
 
-/* Returns -1 on error, and 0 or 1 on success. If -1 or 1 is returned, line has
- * been released and should not be used anymore by the caller. */
-static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
-			   struct osmo_fd *bfd)
+static inline struct osmo_stream_srv *ipaccess_bts_e1i_ts_stream_srv(const struct e1inp_ts *e1i_ts)
 {
+	OSMO_ASSERT(e1i_ts);
+	struct osmo_stream_srv *conn = e1i_ts->driver.ipaccess.fd.data;
+	/* conn may be NULL here, if we are called by user during sign_link_up()
+	 * and hence before we finished to internally move the ipaccess conn under
+	 * the proper final line/ts object.
+	 */
+	return conn;
+}
+
+static int ipaccess_bsc_write_cb(struct e1inp_ts *e1i_ts);
+
+/* Returns -1 on error, and 0 or 1 on success. If -1 or 1 is returned, line has
+ * been released and should not be used anymore by the caller.
+ * msg is always freed upon return.
+ */
+static int ipaccess_bsc_rcvmsg(struct osmo_stream_srv *conn, struct msgb *msg)
+{
+	struct e1inp_ts *e1i_ts = osmo_stream_srv_get_data(conn);
+	struct e1inp_line *line = e1i_ts->line;
+	struct osmo_fd *bfd = &e1i_ts->driver.ipaccess.fd;
 	struct tlv_parsed tlvp;
 	uint8_t msg_type = *(msg->l2h);
 	struct ipaccess_unit unit_data = {};
 	struct e1inp_sign_link *sign_link;
 	char *unitid;
 	int len, ret;
-	struct e1inp_ts *e1i_ts;
-	struct osmo_fsm_inst *ka_fsm;
+	struct osmo_ipa_ka_fsm_inst *ka_fsm;
 
 	/* peek the pong for our keepalive fsm */
-	e1i_ts = ipaccess_line_ts(bfd, line);
 	ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 	if (ka_fsm && msg_type == IPAC_MSGT_PONG)
-		ipa_keepalive_fsm_pong_received(ka_fsm);
+		osmo_ipa_ka_fsm_pong_received(ka_fsm);
 
 	/* Handle IPA PING, PONG and ID_ACK messages. */
 	ret = ipa_ccm_rcvmsg_base(msg, bfd);
@@ -278,7 +247,7 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 		goto err;
 	case 1:
 		/* this is an IPA control message, skip further processing */
-		return 0;
+		goto ret_skip;
 	case 0:
 		/* this is not an IPA control message, continue */
 		break;
@@ -329,17 +298,16 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 				goto err;
 			}
 
-			ipaccess_bsc_keepalive_fsm_alloc(e1i_ts, bfd, "oml_bsc_to_bts");
+			ipaccess_bsc_keepalive_fsm_alloc(e1i_ts, conn, "oml_bsc_to_bts");
 
 		} else if (bfd->priv_nr == E1INP_SIGN_RSL) {
-			struct e1inp_ts *ts;
+			struct e1inp_ts *new_ts;
 			struct osmo_fd *newbfd;
 			struct e1inp_line *new_line;
 			char tcp_stat_name[64];
 
 			sign_link =
-				line->ops->sign_link_up(&unit_data, line,
-							E1INP_SIGN_RSL);
+				line->ops->sign_link_up(&unit_data, line, E1INP_SIGN_RSL);
 			if (sign_link == NULL) {
 				LOGPIL(line, DLINP, LOGL_ERROR, "Unable to set signal link, closing socket.\n");
 				goto err;
@@ -352,37 +320,36 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 			if (new_line == line) {
 				LOGPIL(line, DLINP, LOGL_ERROR, "Fix your BSC, you should use the "
 					"E1 line used by the OML link for your RSL link.\n");
-				return 0;
+				goto ret_skip;
 			}
+
+			/* Attach srv_conn to new line+ts: */
 			e1inp_line_get2(new_line, "ipa_bfd");
-			ts = e1inp_line_ipa_rsl_ts(new_line, unit_data.trx_id);
-			newbfd = &ts->driver.ipaccess.fd;
+			new_ts = e1inp_line_ipa_rsl_ts(new_line, unit_data.trx_id);
+			newbfd = &new_ts->driver.ipaccess.fd;
 			OSMO_ASSERT(newbfd != bfd);
+			osmo_fd_setup(newbfd, bfd->fd, 0, NULL, conn, E1INP_SIGN_RSL + unit_data.trx_id);
+			osmo_stream_srv_set_data(conn, new_ts);
 
-			/* preserve 'newbfd->when' flags potentially set by sign_link_up() */
-			osmo_fd_setup(newbfd, bfd->fd, newbfd->when | bfd->when, bfd->cb,
-				      new_line, E1INP_SIGN_RSL + unit_data.trx_id);
-
-
-			/* now we can release the dummy RSL line (old temporary bfd). */
-			osmo_fd_unregister(bfd);
+			/* Now we can release the old temporary dummy RSL line+ts: */
+			osmo_stats_tcp_osmo_fd_unregister(bfd);
 			bfd->fd = -1;
-			/* bfd->data holds a reference to line, drop it */
-			OSMO_ASSERT(bfd->data == line);
+			/* conn holds a reference to old line, drop it */
+			OSMO_ASSERT(bfd->data == conn);
 			bfd->data = NULL;
 			e1inp_line_put2(line, "ipa_bfd");
 
-			ret = osmo_fd_register(newbfd);
-			if (ret < 0) {
-				LOGPITS(ts, DLINP, LOGL_ERROR, "could not register FD\n");
-				goto err;
-			}
 			snprintf(tcp_stat_name, sizeof(tcp_stat_name), "site.%u.bts.%u.ipa-rsl.%u",
 				unit_data.site_id, unit_data.bts_id, unit_data.trx_id);
 			osmo_stats_tcp_osmo_fd_register(newbfd, tcp_stat_name);
 
-			e1i_ts = ipaccess_line_ts(newbfd, new_line);
-			ipaccess_bsc_keepalive_fsm_alloc(e1i_ts, newbfd, "rsl_bsc_to_bts");
+			ipaccess_bsc_keepalive_fsm_alloc(new_ts, conn, "rsl_bsc_to_bts");
+
+			/* Now that we attached the srv_conn to the e1ts, transmit packets which may been enqueued
+			 * by user during sign_link_up() and which were delayed due to no srv_conn found: */
+			ipaccess_bsc_write_cb(new_ts);
+			/* Conn is not freed here, so we need to free msgb. */
+			msgb_free(msg);
 			return 1;
 		}
 		break;
@@ -390,86 +357,14 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 		LOGPIL(line, DLINP, LOGL_ERROR, "Unknown IPA message type\n");
 		goto err;
 	}
+
+ret_skip:
+	msgb_free(msg);
 	return 0;
 err:
-	if (bfd->fd != -1) {
-		osmo_fd_unregister(bfd);
-		close(bfd->fd);
-		bfd->fd = -1;
-		/* This is a BSC accepted socket, bfd->data holds a reference to line, drop it */
-		OSMO_ASSERT(bfd->data == line);
-		bfd->data = NULL;
-		e1inp_line_put2(line, "ipa_bfd");
-	}
-	return -1;
-}
-
-/* Returns -EBADF if bfd cannot be used by the caller anymore after return. */
-static int handle_ts1_read(struct osmo_fd *bfd)
-{
-	struct e1inp_line *line = bfd->data;
-	unsigned int ts_nr = bfd->priv_nr;
-	struct e1inp_ts *e1i_ts;
-	struct e1inp_sign_link *link;
-	struct ipaccess_head *hh;
-	struct msgb *msg = NULL;
-	int ret, rc;
-
-	e1i_ts = ipaccess_line_ts(bfd, line);
-	ret = ipa_msg_recv_buffered(bfd->fd, &msg, &e1i_ts->pending_msg);
-	if (ret < 0) {
-		if (ret == -EAGAIN)
-			return 0;
-		LOGPITS(e1i_ts, DLINP, LOGL_NOTICE, "Sign link problems, closing socket. Reason: %s\n",
-			strerror(-ret));
-		goto err;
-	} else if (ret == 0) {
-		LOGPITS(e1i_ts, DLINP, LOGL_NOTICE, "Sign link vanished, dead socket\n");
-		goto err;
-	}
-	LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "RX %u: %s\n", ts_nr, osmo_hexdump(msgb_l2(msg), msgb_l2len(msg)));
-
-	hh = (struct ipaccess_head *) msg->data;
-	if (hh->proto == IPAC_PROTO_IPACCESS) {
-		ret = ipaccess_rcvmsg(line, msg, bfd);
-		/* BIG FAT WARNING: bfd might no longer exist here (ret != 0),
-		 * since ipaccess_rcvmsg() might have free'd it !!! */
-		msgb_free(msg);
-		return ret != 0 ? -EBADF : 0;
-	} else if (e1i_ts->type == E1INP_TS_TYPE_NONE) {
-		/* this sign link is not know yet.. complain. */
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Timeslot is not configured.\n");
-		goto err_msg;
-	}
-
-	link = e1inp_lookup_sign_link(e1i_ts, hh->proto, 0);
-	if (!link) {
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "no matching signalling link for hh->proto=0x%02x\n", hh->proto);
-		goto err_msg;
-	}
-	msg->dst = link;
-
-	/* XXX better use e1inp_ts_rx? */
-	if (!e1i_ts->line->ops->sign_link) {
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Fix your application, no action set for signalling messages.\n");
-		goto err_msg;
-	}
-	rc = e1i_ts->line->ops->sign_link(msg);
-	if (rc < 0) {
-		/* Don't close the signalling link if the upper layers report
-		 * an error, that's too strict. BTW, the signalling layer is
-		 * resposible for releasing the message.
-		 */
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Bad signalling message, sign_link returned error: %s.\n",
-			strerror(-rc));
-	}
-
-	return rc;
-err_msg:
 	msgb_free(msg);
-err:
-	ipaccess_drop(bfd, line);
-	return -EBADF;
+	osmo_stream_srv_destroy(conn);
+	return -1;
 }
 
 static void ipaccess_close(struct e1inp_sign_link *sign_link)
@@ -477,12 +372,13 @@ static void ipaccess_close(struct e1inp_sign_link *sign_link)
 	struct e1inp_ts *e1i_ts = sign_link->ts;
 	struct osmo_fd *bfd = &e1i_ts->driver.ipaccess.fd;
 	struct e1inp_line *line = e1i_ts->line;
-	struct osmo_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
+	struct osmo_ipa_ka_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 	struct osmo_stream_cli *cli;
+	struct osmo_stream_srv *conn;
 
 	/* depending on caller the fsm might be dead */
 	if (ka_fsm)
-		ipa_keepalive_fsm_stop(ka_fsm);
+		osmo_ipa_ka_fsm_stop(ka_fsm);
 
 	e1inp_int_snd_event(e1i_ts, sign_link, S_L_INP_TEI_DN);
 	/* the first e1inp_sign_link_destroy call closes the socket. */
@@ -495,19 +391,24 @@ static void ipaccess_close(struct e1inp_sign_link *sign_link)
 		bfd->fd = -1; /* Compatibility with older implementations */
 		break;
 	case E1INP_LINE_R_BSC:
-	default:
-		if (bfd->fd != -1) {
-			osmo_fd_unregister(bfd);
-			close(bfd->fd);
-			bfd->fd = -1;
-			/* If The bfd holds a reference to e1inp_line in ->data (BSC
-			* accepted() sockets), then release it */
-			if (bfd->data == line) {
-				bfd->data = NULL;
-				e1inp_line_put2(line, "ipa_bfd");
-			}
+		/* conn may be NULL here, if we are called by user during sign_link_up()
+		 * callback and hence before we finished to internally move the ipaccess conn
+		 * under the proper final line/ts object:
+		 */
+		conn = ipaccess_bts_e1i_ts_stream_srv(e1i_ts);
+		if (conn) {
+			osmo_stream_srv_destroy(conn);
+		} else {
+			LOGPITS(e1i_ts, DLINP, LOGL_DEBUG,
+				"ipaccess_close() on ts with no srv_conn, "
+				"probably called during sign_link_up() or sign_link_down() user cb.\n");
 		}
 		break;
+	default:
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR,
+			"ipaccess_close() for line with unknown role %d\n",
+			line->ops->cfg.ipa.role);
+		break;
 	}
 }
 
@@ -557,11 +458,66 @@ static int ipaccess_bts_write_cb(struct e1inp_ts *e1i_ts)
 		struct e1inp_sign_link *sign_link = NULL;
 		struct msgb *msg;
 		msg = e1inp_tx_ts(e1i_ts, &sign_link);
+		OSMO_ASSERT(msg);
+		OSMO_ASSERT(sign_link);
 		rc |= ipaccess_bts_send_msg(e1i_ts, sign_link, cli, msg);
 	}
 	return rc;
 }
 
+static int ipaccess_bsc_send_msg(struct e1inp_ts *e1i_ts,
+				 struct e1inp_sign_link *sign_link,
+				 struct osmo_stream_srv *conn,
+				 struct msgb *msg)
+{
+	switch (sign_link->type) {
+	case E1INP_SIGN_OML:
+	case E1INP_SIGN_RSL:
+	case E1INP_SIGN_OSMO:
+		break;
+	default:
+		LOGPITS(e1i_ts, DLMI, LOGL_NOTICE, "Drop Tx msg for unknown sign_link type %d: %s\n",
+			sign_link->type, osmo_hexdump(msg->data, msgb_length(msg)));
+		msgb_free(msg);
+		return -EINVAL;
+	}
+
+	msg->l2h = msg->data;
+	ipa_prepend_header(msg, sign_link->tei);
+
+	LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "TX: %s\n", osmo_hexdump(msg->l2h, msgb_l2len(msg)));
+	osmo_stream_srv_send(conn, msg);
+	return 0;
+}
+
+/* msg was enqueued in sign_link->tx_list.
+ * Pop it from that list, submit it to osmo_stream_srv. */
+static int ipaccess_bsc_write_cb(struct e1inp_ts *e1i_ts)
+{
+	int rc = 0;
+	struct osmo_stream_srv *srv = ipaccess_bts_e1i_ts_stream_srv(e1i_ts);
+
+	/* conn may be NULL here, if we are called by user during sign_link_up()
+	 * ballback and hence before we finished to internally move the ipaccess conn
+	 * under the proper final line/ts object:
+	 */
+	if (!srv) {
+		LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "Delaying Tx on ts with no srv_conn\n");
+		return -ENODEV;
+	}
+
+	/* get the next msg for this timeslot */
+	while (e1i_ts_has_pending_tx_msgs(e1i_ts)) {
+		struct e1inp_sign_link *sign_link = NULL;
+		struct msgb *msg;
+		msg = e1inp_tx_ts(e1i_ts, &sign_link);
+		OSMO_ASSERT(msg);
+		OSMO_ASSERT(sign_link);
+		rc |= ipaccess_bsc_send_msg(e1i_ts, sign_link, srv, msg);
+	}
+	return rc;
+}
+
 static int ts_want_write(struct e1inp_ts *e1i_ts)
 {
 	enum e1inp_line_role role = E1INP_LINE_R_NONE;
@@ -576,118 +532,92 @@ static int ts_want_write(struct e1inp_ts *e1i_ts)
 		/* msg was enqueued in sign_link->tx_list.
 		 * Pop it from that list, submit it to osmo_stream_cli: */
 		return ipaccess_bts_write_cb(e1i_ts);
-	case E1INP_LINE_R_NONE:
 	case E1INP_LINE_R_BSC:
+		/* msg was enqueued in sign_link->tx_list.
+		 * Pop it from that list, submit it to osmo_stream_cli: */
+		return ipaccess_bsc_write_cb(e1i_ts);
+	case E1INP_LINE_R_NONE:
 	default:
-		osmo_fd_write_enable(&e1i_ts->driver.ipaccess.fd);
-		/* ipaccess_fd_cb will be called from main loop and tx the msgb. */
+		LOGPITS(e1i_ts, DLMI, LOGL_ERROR, "want_write for ts with unknown role %d!\n", role);
 		return 0;
 	}
 }
 
-static void timeout_ts1_write(void *data)
+static int ipaccess_bsc_conn_read_cb(struct osmo_stream_srv *conn, int res, struct msgb *msg)
 {
-	struct e1inp_ts *e1i_ts = (struct e1inp_ts *)data;
+	enum ipaccess_proto ipa_proto = osmo_ipa_msgb_cb_proto(msg);
+	struct e1inp_ts *e1i_ts = osmo_stream_srv_get_data(conn);
+	struct e1inp_sign_link *link;
+	int ret, rc;
 
-	/* trigger write of ts1, due to tx delay timer */
-	if (e1i_ts_has_pending_tx_msgs(e1i_ts))
-		ts_want_write(e1i_ts);
-}
-
-static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
-{
-	unsigned int ts_nr = bfd->priv_nr;
-	struct e1inp_ts *e1i_ts;
-	struct e1inp_sign_link *sign_link;
-	struct msgb *msg;
-	int ret;
-
-	e1i_ts = ipaccess_line_ts(bfd, line);
-
-	/* get the next msg for this timeslot */
-	msg = e1inp_tx_ts(e1i_ts, &sign_link);
-	if (!msg) {
-		/* no message after tx delay timer */
-		osmo_fd_write_disable(bfd);
-		return 0;
-	}
-
-	switch (sign_link->type) {
-	case E1INP_SIGN_OML:
-	case E1INP_SIGN_RSL:
-	case E1INP_SIGN_OSMO:
-		break;
-	default:
-		/* leave WRITE flag enabled, come back for more msg */
-		ret = -EINVAL;
-		goto out;
-	}
-
-	msg->l2h = msg->data;
-	ipa_prepend_header(msg, sign_link->tei);
-
-	LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "TX %u: %s\n", ts_nr,
-		osmo_hexdump(msg->l2h, msgb_l2len(msg)));
-
-	ret = send(bfd->fd, msg->data, msg->len, 0);
-	if (ret != msg->len) {
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "failed to send A-bis IPA signalling "
-			"message. Reason: %s\n", strerror(errno));
+	if (res <= 0) {
+		LOGPITS(e1i_ts, DLINP, LOGL_NOTICE, "failed reading from socket: %d\n", res);
 		goto err;
 	}
 
-	/* this is some ancient code that apparently exists to slow down writes towards
-	 * some even more ancient nanoBTS 900 units. See git commit
-	 * d49fc5ae24fc9d44d2b284392ab619cc7a69a876 of openbsc.git (now osmo-bsc.git) */
-	if (e1i_ts->sign.delay) {
-		osmo_fd_write_disable(bfd);
-		/* set tx delay timer for next event */
-		osmo_timer_setup(&e1i_ts->sign.tx_timer, timeout_ts1_write, e1i_ts);
-		osmo_timer_schedule(&e1i_ts->sign.tx_timer, 0, e1i_ts->sign.delay);
-	} else {
-out:
-		if (!e1i_ts_has_pending_tx_msgs(e1i_ts))
-			osmo_fd_write_disable(bfd);
+	LOGPITS(e1i_ts, DLMI, LOGL_DEBUG, "RX: %s\n", osmo_hexdump(msgb_l2(msg), msgb_l2len(msg)));
+
+	if (ipa_proto == IPAC_PROTO_IPACCESS) {
+		ret = ipaccess_bsc_rcvmsg(conn, msg);
+		/* BIG FAT WARNING: bfd might no longer exist here (ret != 0),
+		 * since ipaccess_rcvmsg() might have free'd it !!! */
+		return ret != 0 ? -EBADF : 0;
+	}
+	if (e1i_ts->type == E1INP_TS_TYPE_NONE) {
+		/* this sign link is not know yet.. complain. */
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Timeslot is not configured.\n");
+		goto err;
 	}
-	msgb_free(msg);
-	return ret;
-err:
-	ipaccess_drop(bfd, line);
-	msgb_free(msg);
-	return ret;
-}
 
-static int handle_ts1_write(struct osmo_fd *bfd)
-{
-	struct e1inp_line *line = bfd->data;
+	link = e1inp_lookup_sign_link(e1i_ts, ipa_proto, 0);
+	if (!link) {
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "no matching signalling link for ipa_proto=0x%02x\n", ipa_proto);
+		goto err;
+	}
+	msg->dst = link;
 
-	return __handle_ts1_write(bfd, line);
-}
-
-
-/* callback from select.c in case one of the fd's can be read/written */
-int ipaccess_fd_cb(struct osmo_fd *bfd, unsigned int what)
-{
-	int rc = 0;
-
-	if (what & OSMO_FD_READ)
-		rc = handle_ts1_read(bfd);
-	if (rc != -EBADF && (what & OSMO_FD_WRITE))
-		rc = handle_ts1_write(bfd);
+	/* XXX better use e1inp_ts_rx? */
+	if (!e1i_ts->line->ops->sign_link) {
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Fix your application, no action set for signalling messages.\n");
+		goto err;
+	}
+	rc = e1i_ts->line->ops->sign_link(msg);
+	if (rc < 0) {
+		/* Don't close the signalling link if the upper layers report
+		 * an error, that's too strict. BTW, the signalling layer is
+		 * resposible for releasing the message.
+		 */
+		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "Bad signalling message, sign_link returned error: %s.\n",
+			strerror(-rc));
+	}
 
 	return rc;
+err:
+	msgb_free(msg);
+	osmo_stream_srv_destroy(conn);
+	return 0;
 }
 
-static int ipaccess_line_update(struct e1inp_line *line);
+static int ipaccess_bsc_conn_closed_cb(struct osmo_stream_srv *conn)
+{
+	struct e1inp_ts *e1i_ts = osmo_stream_srv_get_data(conn);
+	struct e1inp_line *line = e1i_ts->line;
+	struct osmo_fd *bfd = &e1i_ts->driver.ipaccess.fd;
+	e1inp_line_get2(line, __func__);
+	e1inp_line_put2(line, "ipa_bfd");
 
-struct e1inp_driver ipaccess_driver = {
-	.name = "ipa",
-	.want_write = ts_want_write,
-	.line_update = ipaccess_line_update,
-	.close = ipaccess_close,
-	.default_delay = 0,
-	.has_keepalive = 1,
-};
+	osmo_stats_tcp_osmo_fd_unregister(bfd);
+	ipaccess_keepalive_fsm_cleanup(e1i_ts);
+
+	osmo_stream_srv_set_data(conn, NULL);
+	bfd->fd = -1;
+	bfd->data = NULL;
+
+	if (line->ops->sign_link_down)
+		line->ops->sign_link_down(line);
+	e1inp_line_put2(line, __func__);
+	return 0;
+}
 
 static void update_fd_settings(struct e1inp_line *line, int fd)
 {
@@ -741,16 +671,19 @@ static void update_fd_settings(struct e1inp_line *line, int fd)
 }
 
 /* callback of the OML listening filedescriptor */
-static int ipaccess_bsc_oml_cb(struct ipa_server_link *link, int fd)
+static int ipaccess_bsc_oml_accept_cb(struct osmo_stream_srv_link *link, int fd)
 {
 	int ret;
 	int i;
+	struct e1inp_line *srv_link_line = osmo_stream_srv_link_get_data(link);
 	struct e1inp_line *line;
+	struct osmo_stream_srv *conn;
 	struct e1inp_ts *e1i_ts;
 	struct osmo_fd *bfd;
+	char conn_name[128];
 
 	/* clone virtual E1 line for this new OML link. */
-	line = e1inp_line_clone(tall_ipa_ctx, link->line, "ipa_bfd");
+	line = e1inp_line_clone(tall_ipa_ctx, srv_link_line, "ipa_bfd");
 	if (line == NULL) {
 		LOGP(DLINP, LOGL_ERROR, "could not clone E1 line\n");
 		return -ENOMEM;
@@ -765,45 +698,47 @@ static int ipaccess_bsc_oml_cb(struct ipa_server_link *link, int fd)
 
 	e1i_ts = e1inp_line_ipa_oml_ts(line);
 
+	snprintf(conn_name, sizeof(conn_name), "ts-%u-%u-oml", line->num, e1i_ts->num);
+	conn = osmo_stream_srv_create2(link, link, fd, e1i_ts);
+	OSMO_ASSERT(conn);
+	osmo_stream_srv_set_name(conn, conn_name);
+	osmo_stream_srv_set_read_cb(conn, ipaccess_bsc_conn_read_cb);
+	osmo_stream_srv_set_closed_cb(conn, ipaccess_bsc_conn_closed_cb);
+	osmo_stream_srv_set_segmentation_cb(conn, osmo_ipa_segmentation_cb);
+
+	/* We use bfd->fd in here for osmo_stats_tcp, and bfd->data to access osmo_stream_srv from e1i_ts. */
 	bfd = &e1i_ts->driver.ipaccess.fd;
-	osmo_fd_setup(bfd, fd, OSMO_FD_READ, ipaccess_fd_cb, line, E1INP_SIGN_OML);
-	ret = osmo_fd_register(bfd);
-	if (ret < 0) {
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "could not register FD\n");
-		goto err_line;
-	}
+	osmo_fd_setup(bfd, fd, 0, NULL, conn, E1INP_SIGN_OML);
 	osmo_stats_tcp_osmo_fd_register(bfd, "ipa-oml");
 
-	update_fd_settings(line, bfd->fd);
+	update_fd_settings(line, fd);
 
 	/* Request ID. FIXME: request LOCATION, HW/SW VErsion, Unit Name, Serno */
-	ret = ipa_ccm_send_id_req(bfd->fd);
+	ret = ipa_ccm_send_id_req(fd);
 	if (ret < 0) {
 		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "could not send ID REQ. Reason: %s\n", strerror(errno));
 		goto err_socket;
 	}
-	return ret;
+	return 0;
 
 err_socket:
-	osmo_fd_unregister(bfd);
-err_line:
-	close(bfd->fd);
-	bfd->fd = -1;
-	bfd->data = NULL;
-	e1inp_line_put2(line, "ipa_bfd");
-	return ret;
+	osmo_stream_srv_destroy(conn);
+	return 0;
 }
 
-static int ipaccess_bsc_rsl_cb(struct ipa_server_link *link, int fd)
+static int ipaccess_bsc_rsl_accept_cb(struct osmo_stream_srv_link *link, int fd)
 {
+	struct e1inp_line *srv_link_line = osmo_stream_srv_link_get_data(link);
 	struct e1inp_line *line;
+	struct osmo_stream_srv *conn;
 	struct e1inp_ts *e1i_ts;
 	struct osmo_fd *bfd;
 	int i, ret;
+	char conn_name[128];
 
 	/* We don't know yet which OML link to associate it with. Thus, we
 	 * allocate a temporary E1 line until we have received ID. */
-	line = e1inp_line_clone(tall_ipa_ctx, link->line, "ipa_bfd");
+	line = e1inp_line_clone(tall_ipa_ctx, srv_link_line, "ipa_bfd");
 	if (line == NULL) {
 		LOGP(DLINP, LOGL_ERROR, "could not clone E1 line\n");
 		return -ENOMEM;
@@ -812,110 +747,37 @@ static int ipaccess_bsc_rsl_cb(struct ipa_server_link *link, int fd)
 	for (i = 0; i < ARRAY_SIZE(line->ts); ++i)
 		line->ts[i].driver.ipaccess.fd.fd = -1;
 
-	/* we need this to initialize this in case to avoid crashes in case
+	/* we need to initialize this in case to avoid crashes in case
 	 * that the socket is closed before we've seen an ID_RESP. */
 	e1inp_ts_config_sign(e1inp_line_ipa_oml_ts(line), line);
 
 	e1i_ts = e1inp_line_ipa_rsl_ts(line, 0);
 
+	snprintf(conn_name, sizeof(conn_name), "ts-%u-%u-rsl", line->num, e1i_ts->num);
+	conn = osmo_stream_srv_create2(link, link, fd, e1i_ts);
+	OSMO_ASSERT(conn);
+	osmo_stream_srv_set_name(conn, conn_name);
+	osmo_stream_srv_set_read_cb(conn, ipaccess_bsc_conn_read_cb);
+	osmo_stream_srv_set_closed_cb(conn, ipaccess_bsc_conn_closed_cb);
+	osmo_stream_srv_set_segmentation_cb(conn, osmo_ipa_segmentation_cb);
+
+	/* We use bfd->fd in here for osmo_stats_tcp, and bfd->data to access osmo_stream_srv from e1i_ts. */
 	bfd = &e1i_ts->driver.ipaccess.fd;
-	osmo_fd_setup(bfd, fd, OSMO_FD_READ, ipaccess_fd_cb, line, E1INP_SIGN_RSL);
-	ret = osmo_fd_register(bfd);
-	if (ret < 0) {
-		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "could not register FD\n");
-		goto err_line;
-	}
+	osmo_fd_setup(bfd, fd, 0, NULL, conn, E1INP_SIGN_RSL);
 	osmo_stats_tcp_osmo_fd_register(bfd, "ipa-rsl");
 
 	/* Request ID. FIXME: request LOCATION, HW/SW VErsion, Unit Name, Serno */
-	ret = ipa_ccm_send_id_req(bfd->fd);
+	ret = ipa_ccm_send_id_req(fd);
 	if (ret < 0) {
 		LOGPITS(e1i_ts, DLINP, LOGL_ERROR, "could not send ID REQ. Reason: %s\n", strerror(errno));
 		goto err_socket;
 	}
-	update_fd_settings(line, bfd->fd);
-	return ret;
+	update_fd_settings(line, fd);
+	return 0;
 
 err_socket:
-	osmo_fd_unregister(bfd);
-err_line:
-	close(bfd->fd);
-	bfd->fd = -1;
-	bfd->data = NULL;
-	e1inp_line_put2(line, "ipa_bfd");
-	return ret;
-}
-
-int ipaccess_bts_handle_ccm(struct ipa_client_conn *link,
-			    struct ipaccess_unit *dev, struct msgb *msg)
-{
-	struct ipaccess_head *hh = (struct ipaccess_head *) msg->data;
-
-	/* special handling for IPA CCM. */
-	if (hh->proto != IPAC_PROTO_IPACCESS)
-		return 0;
-
-	int ret = 0;
-	uint8_t *data = msgb_l2(msg);
-	int len = msgb_l2len(msg);
-	OSMO_ASSERT(len > 0);
-	uint8_t msg_type = *data;
-	/* line might not exist if != bsc||bts */
-	struct e1inp_line *line = link->line;
-
-	/* peek the pong for our keepalive fsm */
-	if (line && msg_type == IPAC_MSGT_PONG) {
-		struct osmo_fsm_inst *ka_fsm = ipaccess_line_ts(link->ofd, line)->driver.ipaccess.ka_fsm;
-		ipa_keepalive_fsm_pong_received(ka_fsm);
-	}
-
-	/* ping, pong and acknowledgment cases. */
-	ret = ipa_ccm_rcvmsg_bts_base(msg, link->ofd);
-	if (ret < 0)
-		goto err;
-
-	/* this is a request for identification from the BSC. */
-	if (msg_type == IPAC_MSGT_ID_GET) {
-		struct msgb *rmsg;
-		/* The ipaccess_unit dev holds generic identity for the whole
-		 * line, hence no trx_id. Patch ipaccess_unit during call to
-		 * ipa_ccm_make_id_resp_from_req() to identify this TRX: */
-		int store_trx_nr = dev->trx_id;
-		if (link->ofd->priv_nr >= E1INP_SIGN_RSL)
-			dev->trx_id = link->ofd->priv_nr - E1INP_SIGN_RSL;
-		else
-			dev->trx_id = 0;
-		LOGP(DLINP, LOGL_NOTICE, "received ID_GET for unit ID %u/%u/%u\n",
-		     dev->site_id, dev->bts_id, dev->trx_id);
-		rmsg = ipa_ccm_make_id_resp_from_req(dev, data + 1, len - 1);
-		dev->trx_id = store_trx_nr;
-		if (!rmsg) {
-			LOGP(DLINP, LOGL_ERROR, "Failed parsing ID_GET message.\n");
-			goto err;
-		}
-
-		ret = ipa_send(link->ofd->fd, rmsg->data, rmsg->len);
-		if (ret != rmsg->len) {
-			LOGP(DLINP, LOGL_ERROR, "cannot send ID_RESP message. Reason: %s\n",
-				strerror(errno));
-			msgb_free(rmsg);
-			goto err;
-		}
-		msgb_free(rmsg);
-
-		/* send ID_ACK. */
-		ret = ipa_ccm_send_id_ack(link->ofd->fd);
-		if (ret <= 0) {
-			LOGP(DLINP, LOGL_ERROR, "cannot send ID_ACK message. Reason: %s\n",
-				strerror(errno));
-			goto err;
-		}
-	}
-	return 1;
-
-err:
-	ipa_client_conn_close(link);
-	return -1;
+	osmo_stream_srv_destroy(conn);
+	return 0;
 }
 
 static struct msgb *ipa_bts_id_ack(void)
@@ -950,12 +812,13 @@ static int _ipaccess_bts_handle_ccm(struct osmo_stream_cli *cli,
 
 	/* peek the pong for our keepalive fsm */
 	if (line && msg_type == IPAC_MSGT_PONG) {
-		struct osmo_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
-		ipa_keepalive_fsm_pong_received(ka_fsm);
+		struct osmo_ipa_ka_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
+		osmo_ipa_ka_fsm_pong_received(ka_fsm);
 	}
 
 	/* ping, pong and acknowledgment cases. */
 	struct osmo_fd tmp_ofd = { .fd = osmo_stream_cli_get_fd(cli) };
+	OSMO_ASSERT(tmp_ofd.fd >= 0);
 	ret = ipa_ccm_rcvmsg_bts_base(msg, &tmp_ofd);
 	if (ret < 0)
 		goto err;
@@ -1067,11 +930,11 @@ static int ipaccess_bts_connect_cb(struct osmo_stream_cli *cli)
 {
 	struct e1inp_ts *e1i_ts = osmo_stream_cli_get_data(cli);
 	struct e1inp_line *line = e1i_ts->line;
-	struct osmo_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
+	struct osmo_ipa_ka_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 
 	update_fd_settings(line, osmo_stream_cli_get_fd(cli));
 	if (ka_fsm && line->ipa_kap)
-		ipa_keepalive_fsm_start(ka_fsm);
+		osmo_ipa_ka_fsm_start(ka_fsm);
 	return 0;
 }
 
@@ -1085,6 +948,7 @@ static int ipaccess_line_update(struct e1inp_line *line)
 {
 	int ret = -ENOENT;
 	struct ipaccess_line *il;
+	char conn_name[128];
 
 	if (!line->driver_data)
 		line->driver_data = talloc_zero(line, struct ipaccess_line);
@@ -1100,38 +964,49 @@ static int ipaccess_line_update(struct e1inp_line *line)
 		/* We only initialize this line once. */
 		if (il->line_already_initialized)
 			return 0;
-		struct ipa_server_link *oml_link, *rsl_link;
+		struct osmo_stream_srv_link *oml_link, *rsl_link;
 		const char *ipa = e1inp_ipa_get_bind_addr();
 
 		LOGPIL(line, DLINP, LOGL_NOTICE, "enabling ipaccess BSC mode on %s "
 		       "with OML %u and RSL %u TCP ports\n", ipa, IPA_TCP_PORT_OML, IPA_TCP_PORT_RSL);
 
-		oml_link = ipa_server_link_create(tall_ipa_ctx, line, ipa,
-						  IPA_TCP_PORT_OML,
-						  ipaccess_bsc_oml_cb, NULL);
-		if (oml_link == NULL) {
-			LOGPIL(line, DLINP, LOGL_ERROR, "cannot create OML BSC link: %s\n", strerror(errno));
-			return -ENOMEM;
-		}
-		oml_link->dscp = g_e1inp_ipaccess_pars.oml.dscp;
-		oml_link->priority = g_e1inp_ipaccess_pars.oml.priority;
-		if (ipa_server_link_open(oml_link) < 0) {
-			LOGPIL(line, DLINP, LOGL_ERROR, "cannot open OML BSC link: %s\n", strerror(errno));
-			ipa_server_link_destroy(oml_link);
+		oml_link = osmo_stream_srv_link_create(tall_ipa_ctx);
+		OSMO_ASSERT(oml_link);
+		snprintf(conn_name, sizeof(conn_name), "ts-%u-oml", line->num);
+		osmo_stream_srv_link_set_name(oml_link, conn_name);
+		osmo_stream_srv_link_set_proto(oml_link, IPPROTO_TCP);
+		osmo_stream_srv_link_set_addr(oml_link, ipa);
+		osmo_stream_srv_link_set_port(oml_link, IPA_TCP_PORT_OML);
+		osmo_stream_srv_link_set_data(oml_link, line);
+		osmo_stream_srv_link_set_nodelay(oml_link, true);
+		osmo_stream_srv_link_set_priority(oml_link, g_e1inp_ipaccess_pars.oml.dscp);
+		osmo_stream_srv_link_set_ip_dscp(oml_link, g_e1inp_ipaccess_pars.oml.priority);
+		osmo_stream_srv_link_set_accept_cb(oml_link, ipaccess_bsc_oml_accept_cb);
+
+		if (osmo_stream_srv_link_open(oml_link)) {
+			LOGPIL(line, DLINP, LOGL_ERROR, "cannot open OML BTS link %s:%u (%s)\n",
+			       ipa, IPA_TCP_PORT_OML, strerror(errno));
+			osmo_stream_srv_link_destroy(oml_link);
 			return -EIO;
 		}
-		rsl_link = ipa_server_link_create(tall_ipa_ctx, line, ipa,
-						  IPA_TCP_PORT_RSL,
-						  ipaccess_bsc_rsl_cb, NULL);
-		if (rsl_link == NULL) {
-			LOGPIL(line, DLINP, LOGL_ERROR, "cannot create RSL BSC link: %s\n", strerror(errno));
-			return -ENOMEM;
-		}
-		rsl_link->dscp = g_e1inp_ipaccess_pars.rsl.dscp;
-		rsl_link->priority = g_e1inp_ipaccess_pars.rsl.priority;
-		if (ipa_server_link_open(rsl_link) < 0) {
-			LOGPIL(line, DLINP, LOGL_ERROR, "cannot open RSL BSC link: %s\n", strerror(errno));
-			ipa_server_link_destroy(rsl_link);
+
+		rsl_link = osmo_stream_srv_link_create(tall_ipa_ctx);
+		OSMO_ASSERT(rsl_link);
+		snprintf(conn_name, sizeof(conn_name), "ts-%u-rsl", line->num);
+		osmo_stream_srv_link_set_name(rsl_link, conn_name);
+		osmo_stream_srv_link_set_proto(rsl_link, IPPROTO_TCP);
+		osmo_stream_srv_link_set_addr(rsl_link, ipa);
+		osmo_stream_srv_link_set_port(rsl_link, IPA_TCP_PORT_RSL);
+		osmo_stream_srv_link_set_data(rsl_link, line);
+		osmo_stream_srv_link_set_nodelay(rsl_link, true);
+		osmo_stream_srv_link_set_priority(rsl_link, g_e1inp_ipaccess_pars.rsl.dscp);
+		osmo_stream_srv_link_set_ip_dscp(rsl_link, g_e1inp_ipaccess_pars.rsl.priority);
+		osmo_stream_srv_link_set_accept_cb(rsl_link, ipaccess_bsc_rsl_accept_cb);
+
+		if (osmo_stream_srv_link_open(rsl_link)) {
+			LOGPIL(line, DLINP, LOGL_ERROR, "cannot open RSL BTS link %s:%u (%s)\n",
+			       ipa, IPA_TCP_PORT_RSL, strerror(errno));
+			osmo_stream_srv_link_destroy(rsl_link);
 			return -EIO;
 		}
 		ret = 0;
@@ -1293,12 +1168,21 @@ int e1inp_ipa_bts_rsl_close_n(struct e1inp_line *line, uint8_t trx_nr)
 
 	cli = il->ipa_cli[1 + trx_nr];
 	if (cli != NULL) {
-		osmo_stream_cli_destroy(cli);
 		il->ipa_cli[1 + trx_nr] = NULL;
+		osmo_stream_cli_destroy(cli);
 	}
 	return 0;
 }
 
+static struct e1inp_driver ipaccess_driver = {
+	.name = "ipa",
+	.want_write = ts_want_write,
+	.line_update = ipaccess_line_update,
+	.close = ipaccess_close,
+	.default_delay = 0,
+	.has_keepalive = 1,
+};
+
 void e1inp_ipaccess_init(void)
 {
 	tall_ipa_ctx = talloc_named_const(libosmo_abis_ctx, 1, "ipa");
